

// Loading Overlay Control
function showLoading() {
  document.getElementById("loadingOverlay").style["display"] = "";
}

function hideLoading() {
  document.getElementById("loadingOverlay").style["display"] = "none";
}